/* dg_spline.h -- polynomial splines defined on irregular dyadic grids */
/* Last edited on 2004-08-17 23:50:30 by stolfi */

#ifndef dg_spline_H
#define dg_spline_H

#include <dg_tree.h>
#include <js.h>
#include <vec.h>
#include <vec.h>

/* POLYNOMIAL SPLINES */

/* Let {G} be an arbitrary partition of some region {D} of {R^d} into
  pairwise disjoint cells. A `(polynomial) spline' for {G} is a
  function {f} from {R^d} to {R} such that {f(x)} is zero outside {D}
  and, for each cell {c} of {G}, the restriction {(f|c)(x)} is a
  polynomial on the {d} coordinates {x[0..d-1]} of the argument.
  
  Each partial function {f|c} is called a `piece' of the spline. The
  `degree' of the spline is the maximum exponent of any coordinate in
  any its pieces.  The set of all splines of degree {g} on a given 
  grid {G} constitute a linear vector space, {S^g(G)}.

  The procedures in this section are concerned with the sub-space
  {S^1_0(G)} of {S^1(G)}, consisting of the *continuous* splines of degree 1 
  for some finite *dyadic* grid {G}.  Note that since the union of all cells
  of {G} is the root cell {C_0}, any such spline is by definition zero outside 
  {G}, and (by continuity) also on the boundary of {G}.

  A polynomial of degree 1 is a multi-linear function - namely, a
  function which is affine (degree 1) along any axis. Note that if
  such a function is zero at two points of an axis-aligned line
  segment, it is zero on the whole segment. Thus, in particular, if a
  spline {f} from {S^1_0(G)} is zero at the two endpoints of some edge
  {e} of the grid, it is zero along the entire edge {e}. If {f} is
  zero on two opposite edges of some face {H} of the grid, it is zero
  on the whole face {H}. And so on. */ 

/* TENT FUNCTIONS */

/* A `tent function' is a spline {t} in {S^1_0(Q)}, where {Q}
  is a grid consisting of {2^d} axis-aligned boxes {b[0..2^d-1]}, of the same
  shape and size, that share a common corner {p}. By definition,
  within each of these boxes (the `bricks' of {t}), the tent function
  is a single non-zero multilinear function, which decreases from some
  central value {t(p)} to zero on the boundary of the tent's support {B} 
  (the union of its bricks).
  
  A `unit tent function' is a tent function that has value 1 at the
  central point. Observe that all the tent functions with the same
  set of bricks {b[i]} are multiples of the corresponding 
  unit tent function.
  
  A `dyadic tent' is a tent funtion whose bricks are cells of
  the dyadic cell tree. */
  
typedef struct dg_Tent  /* A unit dyadic tent function. */ 
  { dg_Index cx;
    dg_Rank r;
  } dg_Tent;
  /* A diyadic tent function is completely determined by the center
    point {c} and the rank {r} of its bricks. The center {c}, in turn,
    is uniquely identified by the index of the the cell of which {cx}
    is the center [[??? Or the index of lowest-numbred brick ???]].
    
    Thus, for example, the largest possible unit dyadic tent,
    whose support is the entire root cell, spans the {2^d} bricks
    of rank {d}, whose indices are {2^d .. 2^{d+1}-1}; thus
    it can be represented by the pair { cx = 1, r = d }
    [[??? or {cx = 2^d, r = d} ???]]. */

typedef struct dg_Tent_pair  /* For pairs of tents which have cells in comon. */ 
  { dg_Index cx[2];
    dg_Rank r[2];
  } dg_Tent_pair;
 
/* TENT FUNCTION EVALUATION */

/* The evaluation of a unit tent function {t} at a point {p} of {R^d}
  is best divided three steps. First, we compute the
  `support-relative' coordinates {x} of {p}, by the formula {x[i] =
  (p[r+i]-c[r+i])/h[r+i]}, where {r} is the rank of the bricks, {c} is
  the center of the tent's support, and {h[j]} is the extent along
  axis {j} of the bricks of {t}. (All indices are taken modulo {d}.)
  Note that the support-relative coordinate {x[0]} is the relative
  position of {p} along the longest axis of the bricks. 
  
  Then, if any of the {x[i]} lies outside the range {[-1 _ +1]}, the
  value of {t(p)} is zero.  Otherwise, the value is simply the product
  of {1 - |x[i]|}, for all {i}.
  
  In practice, the grid may be deformed by a complicated shape
  function {f}, which needs to be inverted before computing {t(p)}. 
  So, part of the computation of {x} is probbaly best merged 
  with the computation of {f^{-1}(p)}. */

double dg_tent_eval(dg_Dim d, double *c, dg_Rank r, double *x);
  /* Evaluates the {d}-dimensional tent function of rank {r} centered
    on the point {c}, at the point whose support-relative coordinates
    are {x[0..d-1]}. */
 
/* MINIMAL DYADIC TENTS IN A FINITE DYADIC GRID */

/* A dyadic tent {t} is said to be `compatible' with a grid {G} if each
  of its bricks is the union of one or more whole cells of {G}. A
  compatible tent {t} is `minimal' if there is no other dyadic tent
  function compatible {G} whose support is properly contained in that
  of {t}.
  
  If {G} is a finite dyadic grid, it is easy to see that a dyadic tent
  {t} is minimal if, and only if, all its bricks are cells (leaf or
  non-leaf) of {G}, of the same rank {r}, and at least one of them is
  a leaf cell. It then follows that the support {B} of {t} is
  congruent to a dyadic cell of rank {r-d}, translated so that it is
  centered at the common vertex {p} of the bricks. (Note that {B}
  itself may not be a cell of the dyadic cell tree; it may it may
  straddle two or more such cells.)
  
  It is conjectured that, for any dyadic grid {G}, the set of all the
  unit minimal tent functions of {G} constitute a basis for the space
  {S^1_0(G)}. */
    
typedef struct dg_Tent_vec_t { nat nel; dg_Tent *el; } dg_Tent_vec_t;
  /* A list of minimal unit tents. */

typedef struct dg_Tent_pair_vec_t { nat nel; dg_Tent_pair *el; } dg_Tent_pair_vec_t;
  /* A list of minimal tent pairs. */

dg_Tent_vec_t dg_minimal_tent_basis(dg_Tree *t);
  /* Returns a list of all minimal unit tent functions in the 
    space {S^1_0(G)}, where {G} is the finite dyadic cell grid
    described by the tree {t}. */

dg_Tent_vec_t dg_maximal_tent_basis(dg_Tree *t);
  /* Instead of looking for the tents of minimal support, we can 
    look for a basis of {S^1_0(G)} consisting of tents
    with largest possible suport. This procedure returns such a 
    basis. */

dg_Tent_pair_vec_t dg_minimal_tent_pairs(dg_Tree *t);

/* Manipulating vectors of {dg_Tent}: */

bool_t dg_common_cells_found(dg_Rank *r, dg_Node ***C, int *num_cells);

dg_Tent_vec_t dg_tent_vec_new(nat nel);

#define dg_tent_vec_expand(nv,index) \
  vec_expand(vec_cast_ref(nv), index, sizeof(dg_Tent))
  /* Makes sure that the vector {nv} has element {nv.el[index]}.
    Reallocates {nv.el} if necessary, to have at least {index+1}
    elements -- but possibly more. */

#define dg_tent_vec_trim(nv,nel) \
  vec_trim(vec_cast_ref(nv), nel, sizeof(dg_Tent))
  /* Makes sure that the vector {nv} has exactly {nel} elements.
    Reallocates {nv.el} if necessary. */

/* Manipulating vectors of {dg_Tent_pair}: */

dg_Tent_pair_vec_t dg_tent_pair_vec_new(nat nel);

#define dg_tent_pair_vec_expand(nv,index) \
  vec_expand(vec_cast_ref(nv), index, sizeof(dg_Tent_pair))
  /* Makes sure that the vector {nv} has element {nv.el[index]}.
    Reallocates {nv.el} if necessary, to have at least {index+1}
    elements -- but possibly more. */

#define dg_tent_pair_vec_trim(nv,nel) \
  vec_trim(vec_cast_ref(nv), nel, sizeof(dg_Tent_pair))
  /* Makes sure that the vector {nv} has exactly {nel} elements.
    Reallocates {nv.el} if necessary. */

#endif
