/* See dg_pulse.h */
/* Last edited on 2005-08-10 17:23:08 by stolfi */

#include <dg_grid.h>
#include <dg_pulse.h>
#include <dg_pulse_B.h>
#include <dg_pulse_H.h>
#include <dg_pulse_N.h>
#include <bz_basic.h>
#include <affirm.h>
#include <math.h>

/* IMPLS */

char dg_pulse_kind_to_char(dg_pulse_kind_t pkind)
  { switch(pkind)
      { case dg_PK_B: 
          return 'B';
        case dg_PK_H: 
          return 'H';
        case dg_PK_N: 
          return 'N';
        default:
          affirm(FALSE, "unknown pulse kind");
          return '?'; /* To pacify the compiler. */
      }
  }  

dg_pulse_family_t dg_pulse_family(dg_pulse_kind_t pkind, dg_cont_t c, dg_degree_t g)
  { unsigned int nmp;
    switch(pkind)
      { case dg_PK_B: 
          nmp = dg_pulse_B_num_mothers(c, g); break;
        case dg_PK_H: 
          nmp = dg_pulse_H_num_mothers(c, g); break;
        case dg_PK_N: 
          nmp = dg_pulse_N_num_mothers(c, g); break;
        default:
          affirm(FALSE, "unknown pulse kind");
          nmp = 0;
      }
    return (dg_pulse_family_t){ pkind, c, g, nmp }; 
  }

dg_pulse_t dg_pulse
  ( dg_pulse_family_t *fam, 
    dg_pulse_mother_index_t m, 
    dg_grid_size_t gsz, 
    dg_grid_pos_t pos
  )
  { dg_grid_size_t msz = dg_pulse_mother_supp_count(fam, m);
    return (dg_pulse_t){ m, msz, gsz, pos };
  }

void dg_pulse_eval
  ( dg_pulse_family_t *fam,
    dg_pulse_t *p,
    interval_t *R,
    double x,
    dg_degree_t ord,
    double *f
  )
  { /* Scale argument from {R} to {[0 _ gsz)} and shift it from {pos} to 0: */
    double scale = (double)p->gsz, y = x;
    if (R != NULL) 
      { double xlo = (R == NULL ? 0 : LO(*R));
        double xhi = (R == NULL ? 1 : HI(*R));
        scale /= xhi - xlo;  y -= xlo;
      }
    y = y*scale - p->pos;
    
    /* Evaluate the pulse and derivatives w.r.t. {y}: */
    dg_pulse_eval_cell_relative(fam, p->m, p->gsz, y, ord, f);

    if ((ord > 0) && (scale != 1))
      { /* Adjust derivatives to compensate for scaling: */
        double fac = scale;
        int k;
        for (k = 1; k <= ord; k++) { f[k] *= fac; fac *= scale; }
      }
  }

void dg_pulse_eval_cell_relative
  ( dg_pulse_family_t *fam,
    dg_pulse_mother_index_t m, 
    dg_grid_size_t gsz,
    double x,
    dg_cont_t ord,
    double *f
  )
  { affirm(fam->g >= 0, "negative degree");
    affirm(fam->g <= dg_PULSE_MAX_DEGREE, "excessive degree");
    affirm(ord <= dg_PULSE_MAX_DIFF_ORD, "diff order too high");
    
    /* Reduce {x} to {[0 _ gsz)} modulo {gsz}: */
    double z = x - floor(x/gsz)*gsz;
    while (z >= gsz) { z -= gsz; }
    while (z < 0) { z += gsz; }
    
    /* Clear out the result vector: */
    int k;
    for (k = 0; k <= ord; k++) { f[k] = 0; }
    
    /* Discover the high endpoint {hi} of mother pulse's support: */
    dg_grid_pos_t hi = dg_pulse_mother_supp_count(fam, m);

    /* Add mother's values for all {z} in {(0 _ hi)} congruent to {x} mod {gsz}: */
    if (z == 0) { z += gsz; } 
    double w[dg_PULSE_MAX_DEGREE + 1];
    while (z < hi) 
      { dg_pulse_mother_eval(fam, m, z, ord, w);
        for (k = 0; k <= ord; k++) { f[k] += w[k]; }
        z += gsz;
      }
  }

dg_grid_size_t dg_pulse_max_supp_count ( dg_pulse_family_t *fam )
  {
    int m, nsmax = 0;
    for (m = 0; m < fam->nmp; m++)
      { int nsm = dg_pulse_mother_supp_count(fam, m);
        if (nsm > nsmax) { nsmax = nsm; }
      }
    return nsmax;
  }

void dg_pulse_to_bezier
  ( dg_pulse_family_t *fam,
    dg_pulse_t *p, 
    dg_grid_pos_t ix,
    double *bz
  )
  { affirm(fam->g >= 0, "negative degree");
    affirm(fam->g <= dg_PULSE_MAX_DEGREE, "excessive degree");
    
    /* Shift pulse from {pos} to 0: */
    ix -= p->pos;
    
    /* Compute {z} = {ix} reduced to {0..p->gsz-1} modulo {p->gsz}: */
    int z = ix % p->gsz;
    while (z < 0) { z += p->gsz; }
    while (z >= p->gsz) { z -= p->gsz; };
    
    /* Clear out the Bzier coeff vector: */
    int k;
    for (k = 0; k <= fam->g; k++) { bz[k] = 0; }

    /* Add mother's coeffs for all {z} in {0..p->msz-1} congruent to {ix} mod {gsz}: */
    double wbz[dg_PULSE_MAX_DEGREE + 1];
    while (z < p->msz) 
      { dg_pulse_mother_to_bezier(fam, p->m, z, wbz);
        for (k = 0; k <= fam->g; k++) { bz[k] += wbz[k]; }
        z += p->gsz;
      }
  }

/* STANDARD MOTHER PULSES */

dg_grid_size_t dg_pulse_mother_supp_count
  ( dg_pulse_family_t *fam, 
    dg_pulse_mother_index_t m
  )
  { switch(fam->pkind)
      { case dg_PK_B: 
          return dg_pulse_B_mother_supp_count(fam->c, fam->g, m);
        case dg_PK_H: 
          return dg_pulse_H_mother_supp_count(fam->c, fam->g, m);
        case dg_PK_N: 
          return dg_pulse_N_mother_supp_count(fam->c, fam->g, m);
        default:
          affirm(FALSE, "unknown pulse kind");
          return 0.0; /* To keep the compiler happy. */
      }
  }

void dg_pulse_mother_eval
  ( dg_pulse_family_t *fam, 
    dg_pulse_mother_index_t m, 
    double z,
    dg_cont_t ord,
    double *w
  )
  { affirm(fam->g >= 0, "negative degree");
    affirm(fam->g <= dg_PULSE_MAX_DEGREE, "excessive degree");
    affirm(ord <= dg_PULSE_MAX_DIFF_ORD, "diff order too high");
    /* Get Bzier coeffs for interval containing {z}: */
    double bz[dg_PULSE_MAX_DEGREE + 1];
    int k = (int) floor(z); 
    dg_pulse_mother_to_bezier(fam, m, k, bz);
    /* Apply DeCasteljau's algorithm: */
    bz_eval(fam->g, bz, 0.0, 1.0, z - k, ord, w);
  }

void dg_pulse_mother_to_bezier
  ( dg_pulse_family_t *fam, 
    dg_pulse_mother_index_t m,
    dg_grid_pos_t ix,
    double *bz
  )
  { affirm(fam->g >= 0, "negative degree");
    affirm(fam->g <= dg_PULSE_MAX_DEGREE, "excessive degree");
    switch(fam->pkind)
      { case dg_PK_B: 
          dg_pulse_B_mother_to_bezier(fam->c, fam->g, m, ix, bz);
          return;
        case dg_PK_H: 
          dg_pulse_H_mother_to_bezier(fam->c, fam->g, m, ix, bz);
          return;
        case dg_PK_N: 
          dg_pulse_N_mother_to_bezier(fam->c, fam->g, m, ix, bz);
          return;
        default:
          affirm(FALSE, "unknown pulse kind");
      }
  }

void dg_pulse_print
  ( FILE *wr, 
    dg_pulse_family_t *fam, 
    dg_pulse_t *p
  )
  { /* Pulse kind: */
    fprintf(wr, "%c", dg_pulse_kind_to_char(fam->pkind));
    /* Continuity and degree: */
    fprintf(wr, "_%d^%d", fam->c, fam->g);
    /* Mother pulse index: */
    if ((fam->nmp != 1) || (p->m != 0))
      { fprintf(wr, "[");
        fprintf(wr, "%d", p->m);
        if (p->m < 0) { fprintf(wr, "!?"); }
        if (p->m >= fam->nmp) { fprintf(wr, "!?(%d)", fam->nmp); }
        fprintf(wr, "]");
      }
    /* Shift and grid size: */
    dg_grid_size_t psz = (p->msz < p->gsz ? p->msz : p->gsz);
    fprintf(wr, ":%lld..%lld/%lld", p->pos, p->pos+psz-1, p->gsz);
  }
