/* H-pulses - Hermite-like univariate finite elements. */
/* Last edited on 2005-04-22 10:12:46 by stolfi */

#ifndef dg_pulse_H_H
#define dg_pulse_H_H

#include <dg_grid.h>
#include <bz_basic.h>
#include <dg_pulse.h>
#include <js.h>

/* 
  H-PULSES
  
  The dyadic H-pulses are defined only for {c\geq 0}, {g = 2*c+1}.
  
  There are exactly {c+1} mother H-pulses {w^g_c[m]}. Their support
  is the interval {(0 _ 2)}. The mother pulse of index
  {m} has all derivatives up to order {c} equal to zero at integer
  {z} values, except for the derivative of order {m} at {z = 1},
  which is equal to {choose(g,m)}.
  
  The mother pulse is symmetric or anti-symmetric around {z = 1}, 
  depending on the  parity of {m}; so, for {z} in {(0,1)}, 

    { w^g_c[m](z) = (-1)^m w^g_c[m](2-z) }
  
  */

unsigned int dg_pulse_H_num_mothers (dg_cont_t c, dg_degree_t g);
  /* Returns the number of mother H-pulses with continuity {c}
    and degree {g}; namely, {c + 1} if {c >= 0} and 
    {g = 2*c + 1}, and 0 otherwise. */

dg_grid_size_t dg_pulse_H_mother_supp_count
  ( dg_cont_t c,
    dg_degree_t g,
    dg_pulse_mother_index_t m
  ); 
  /* Returns the number of unit-grid cells that comprise the support
    of the mother H-pulse with given {c,g,m}. Fails if such H-pulses
    aren't defined or aren't supported by this module. */

void dg_pulse_H_mother_to_bezier
  ( dg_cont_t c, 
    dg_degree_t g,
    dg_pulse_mother_index_t m,
    dg_grid_pos_t x,
    double *bz
  );
  /* Stores into {bz[0..2*c+1]} the Bzier coefficients of the mother
    H-pulse with continuity {c}, degree {g}, and mother index {m}, for
    the interval {(x _ x+1)}. Fails if such pulse is not defined or
    supported by this module. The coefficients are nonzero only if {x}
    is 0 or 1. */

#endif
