/* Polynomial splines defined on irregular dyadic grids */
/* Last edited on 2005-06-05 20:08:43 by stolfi */

#ifndef dg_tent_basis_H
#define dg_tent_basis_H

#include <dg_grid.h>
#include <bz_basic.h>
#include <dg_tent.h>
#include <dg_tree.h>
#include <vec.h>
#include <js.h>

/*
  DYADIC POLYNOMIAL SPLINE SPACES
  
  For a finite grid {G} and given continuity and degree vectors
  {c[0..d-1]} and {g[0..d-1]}, the space {\S^d[c,g][G]} consists of
  all polynomial splines on {G} whose pieces have degree at most
  {g[i]} on coordinate {x[i]}, and which have continuity {c[i]} along
  any line parallel to axis {i}.

  TENT BASES FOR CONTINUOUS SPLINES
  
  Let {pkind[0..d-1]} be an arbitrary vector of pulse kinds. If {G} is
  a finite dyiadic grid, the space {\S^d[c,g][G]} admits a basis {\CB}
  where  all elements are tents with parameters {pkind[],c[],g[]}.  
  We say that {\CB} is /a basis of the kind {pkind[]}/.
  
  FROM TEMPLATES TO BASES
  
  To build a basis for {\S^d[c,g][G]} of the kind
  {pkind[0..d-1]}, we must find a set of tents with parameters
  {pkind[],c[],g[]}, whose supporting cells are all cells of 
  {G}.  Thus, for each vector of pulse indices {pix[0..d-1]},
  which is compatible with those parameters, we must 
  find a set {\X} of cells that fit the template {psz[0..d-1]}, where
  
    {psz[i] = dg_pulse_mother_supp_size(pkind[i],c[i],g[i],pix[i])}
    
  Then, for every such cell {C} with rank {r} and position {pos[]}, 
  we take the tent function with parameters {pkind,c,g,pix,r,pos}.
  
  ELIMINATING LINEAR DEPENDENCIES
  
  The set of tents described above may not be linearly independent.
  In particular, let {t} be a tent with parameters {pkind,c,g,pix}
  supported on a cell pack {P}. ???
  
  ??? If the support size vector is the same for all pulse indices and
  at least 2 along all axes it suffices to take the packs which
  contain at least one leaf cell. ???.
  ??? [Seems true for S-pulses and H-pulses, not sure for B-pulses
  with {g > 2*c+1}]. ???  If the size along
  any axis is 1 and {c >= 0} then one must take also non-leaf packs. ??? */
  
dg_tent_vec_t dg_tent_basis_get
  ( dg_dim_t d, 
    dg_pulse_family_t fam[], 
    dg_tree_t G
  );
  /* Returns a set of {d}-dimensional tents, consisting 
    of all possible products of pulses from family {fam[i]} along each axis
    {i} in {0..d}. */

/*
  COMPLETE VERTICES

  In particular, a rank-{r} cell {C} of {G} fits the template
  {(2,2,..2)} if and only if every rank-{r} dyadic cell that is
  incident to its upper corner {v} is a cell of {G}.  Thus, for 
  combinations of {pkind,c,g} that yield that template, we
  need to locate the /complete vertices/ of {G} -- namely 
  those which are incident to {2^d} cells (counting repetitions).
 
  INFERIOR AND SUPERIOR TENT BASES
  
  ??? Assuming {g \geq 2c + 1}, a basis for {S(P^g_c,L)} is a list of all
  tent functions {W^g_c[e]} centered at all minimal valid loci of
  {G}. This is the /inferior tent basis/ for {S(P^g_c,L)}.
  
  ??? In general, a dyadic face {f} corresponds to several loci in a
  finite grid. To obtain the /superior tent basis/ for {S(P^g_c,L)},
  we replace each minimal valid locus {E} by the maximal valid locus
  {E'} of same dimension and same item, but maximal (instead of minimal) 
  star. */

#endif
